#!/usr/bin/env python3
#
#    util.py
#
# Copyright (c) 2025 Alicia González Martínez
#
################################################

from typing import TypedDict
from argparse import ArgumentTypeError

from .models import Index, Block


class TextArgs(TypedDict, total=False):
    blocks: bool
    no_lat: bool
    no_ara: bool
    no_graph: bool
    no_arch: bool


def _parse_index(index: str, default: int) -> Index:
    """ convert string containing an index into an Index object.

    Args:
        index: index to parse.
        default: default value if not included in index.

    Return:
        splited index.

    Raise:
        ValueError: if index is ill-formed.

    """
    try:
        indexes = iter(index.split(":"))
        sura, verse, word, block = (next(indexes, None) for _ in range(4))
        return Index(
            sura=int(sura) if sura else default,
            verse=int(verse) if verse else default,
            word=int(word) if word else default,
            block=int(block) if block else default
        )
    except ValueError:
        raise


def parse_quran_range(arg: str) -> tuple[Index, Index]:
    """ Parse index range contained in arg. Complete format:

        "sura,verse,word,block-sura,verse,word,block"

        All individual indexes are optional

    Args:
        arg: string containing the quranic index range or single index.

    Return:
        range of indexes.

    Raise:
        ArgumentTypeError: if arg does not follow the expected format.

    """
    if "-" in arg:
        ini, _, end = arg.partition("-")
    else:
        ini = end = arg

    try:
        return (
            _parse_index(index=ini if ini else "", default=1),
            _parse_index(index=end if end else "", default=-1)
        )

    except ValueError:
        raise ArgumentTypeError(
            "argument format must be sura:verse:word:block-sura:verse:word:block, eg. 2:3-2:10:2"
        )

def group_idx(block: Block) -> tuple[int, int, int]:
    """
    """
    idx = block.index
    return (idx.sura, idx.verse, idx.word)




#!/usr/bin/env python3
#
#    mushaf.py
#
# retrieve Quranic text
#
#  __  __           _            __ 
# |  \/  |_   _ ___| |__   __ _ / _|
# | |\/| | | | / __| '_ \ / _` | |_ 
# | |  | | |_| \__ \ | | | (_| |  _|
# |_|  |_|\__,_|___/_| |_|\__,_|_|  
#
# Copyright (c) 2025 Alicia González Martínez
#
################################################

import sys
import orjson as json
import importlib.resources as pkg_resources
from itertools import groupby
from typing import Generator

from .models import Source, Index, Block
from .util import TextArgs, group_idx


def _extract_seq(
    data: dict,
    ini_ind: Index,
    end_ind: Index
    ) -> Generator[Block, None, None]:
    """ Extracts the sequece of Quranic blocks indicated in index range ind.

    Args:
        data: Quran data
        ind: index range.

    Yield:
        Representations of Quranic token together with its index.

    """
    ini_sura = ini_ind.sura
    ini_verse = ini_ind.verse
    ini_word = ini_ind.word
    ini_block = ini_ind.block

    end_sura = end_ind.sura
    end_verse = end_ind.verse
    end_word = end_ind.word
    end_block = end_ind.block

    for isura in range(ini_sura, end_sura+1):
        for iverse in range(len(data["indexes"][isura])):
            if isura == ini_sura and iverse < ini_verse:
                continue
            if isura == end_sura and iverse > end_verse:
                return
            for iword in range(len(data["indexes"][isura][iverse])):
                if isura == ini_sura and iverse == ini_verse and iword < ini_word:
                    continue
                if isura == end_sura and iverse == end_verse and iword > end_word:
                    return
                for iblock in range(len(data["indexes"][isura][iverse][iword])):
                    if isura == ini_sura and iverse == ini_verse and \
                        iword == ini_word and iblock < ini_block:
                        continue
                    if isura == end_sura and iverse == end_verse and \
                        iword == end_word and iblock > end_block:
                        return

                    block_index = data["indexes"][isura][iverse][iword][iblock]

                    graph_ar, graph_lt, arch_ar, arch_lt = data["blocks"][block_index]
                    
                    yield Block(
                        grapheme_ar=graph_ar,
                        grapheme_lt=graph_lt,
                        archigrapheme_ar=arch_ar, 
                        archigrapheme_lt=arch_lt,
                        index=Index(
                            sura=isura+1,
                            verse=iverse+1,
                            word=iword+1,
                            block=iblock+1
                        )
                    )


def _correct_out_of_bounds(ind: Index, data: dict) -> None:
    """ Adjust all indexes in ind so that none of them is out of bounds.

    ind is still 1-based index

    Args:
       ind: index to correct.
       data: data containig whole Quran structure.


    """
    if ind.sura > len(data["indexes"]):
        print(
            f"Warning! sura {ind.sura} is out of bounds. "
            f"We set it to last sura, i.e. {len(data['indexes'])} ",
            file=sys.stderr
        )
        ind.sura = len(data["indexes"])

    if ind.verse > len(data["indexes"][ind.sura-1]):
        print(
            f"Warning! verse {ind.verse} is out of bounds. "
            f"We set it to last verse in sura, i.e. {len(data['indexes'][ind.sura-1])} ",
            file=sys.stderr
        )
        ind.verse = len(data["indexes"][ind.sura-1])
    
    if ind.word > len(data["indexes"][ind.sura-1][ind.verse-1]):
        print(
            f"Warning! word {ind.word} is out of bounds. "
            f"We set it to last word in verse, i.e. {len(data['indexes'][ind.sura-1][ind.verse-1])} ",
            file=sys.stderr
        )
        ind.word = len(data["indexes"][ind.sura-1][ind.verse-1])
    
    if ind.block > len(data["indexes"][ind.sura-1][ind.verse-1][ind.word-1]):
        print(
            f"Warning! block {ind.block} is out of bounds. "
            f"We set it to last block in word, i.e. {len(data['indexes'][ind.sura-1][ind.verse-1][ind.word-1])} ",
            file=sys.stderr
        )
        ind.block = len(data["indexes"][ind.sura-1][ind.verse-1][ind.word-1])


def get_text(
    ini_index: Index | tuple[int, int, int, int] = (1, 1, 1, 1),
    end_index: Index | tuple[int, int, int, int] = (-1, -1, -1, -1),
    source: Source | str = Source.TANZIL_SIMPLE,
    args: TextArgs = {}
    ) -> Generator[tuple[str, ...], None, None]:
    """ Get all tokens corresponding to the indicated Quranic index ranges.

    Unless indicated with args parameters, each token is given in four shapes:
    (1) Arabic graphemic representation
    (2) Latin graphemic representation
    (3) Arabic archgraphemic representation
    (4) Latin graphemic representation

    Args:
        index: initial and end indexes
        source: Quranic encoding to retrieve:
            (1) "tanzil-simple"
            (2) "tanzil-uthmani"
            (3) "decotype"
        args: parameters to configure desired output 
            {
                "blocks": bool,    # display tokens in blocks instead of words
                "no_lat": bool,    # do not show representations in Latin script
                "no_ara": bool,    # do not show representations in Arabic script
                "no_graph": bool,  # do not show graphemic representations 
                "no_arch": bool,   # do not show archigraphemic representations
            }
    Yield:
        Representations of Quranic token together with its index.

    """
    default_args: TextArgs = {
        "blocks": False,
        "no_lat": False,
        "no_ara": False,
        "no_graph": False,
        "no_arch": False
    }
    args = {**default_args, **args}

    if isinstance(source, str):
        source = Source.from_str(source)

    quran = source.get_file()

    quran_path = pkg_resources.files(__package__).joinpath(quran)
    
    if not quran_path.is_file():
        raise FileNotFoundError("Decotype Quran is private.")

    with quran_path.open("rb") as fp:
        data = json.loads(fp.read())

    if isinstance(ini_index, tuple):
        ini_index = Index.from_tuple(ini_index)

    if isinstance(end_index, tuple):
        end_index = Index.from_tuple(end_index)

    ini = ini_index
    end = end_index

    _correct_out_of_bounds(ini, data)
    _correct_out_of_bounds(end, data)

    # correct -1 end indexes
    if end.sura == -1:
        end.sura = len(data["indexes"])
    
    if end.verse == -1:
        end.verse = len(data["indexes"][end.sura-1])

    if end.word == -1:
        end.word = len(data["indexes"][end.sura-1][end.verse-1])

    if end.block == -1:
        end.block = len(data["indexes"][end.sura-1][end.verse-1][end.word-1])

    ini.to_zero_index()
    end.to_zero_index()

    sequence = _extract_seq(data, ini, end)

    if not args["blocks"]:
        
        seq = ((k, list(g)) for k, g in groupby(sequence, key=group_idx))
        
        sequence = (
            Block(
                grapheme_ar="".join(b.grapheme_ar for b in blocks_group),
                grapheme_lt="".join(b.grapheme_lt for b in blocks_group),
                archigrapheme_ar="".join(b.archigrapheme_ar for b in blocks_group),
                archigrapheme_lt="".join(b.archigrapheme_lt for b in blocks_group),
                index=Index(
                    sura=ind[0],
                    verse=ind[1],
                    word=ind[2],
                    block=None
                )
        ) for ind, blocks_group in seq)

    res: tuple[str, ...]

    for block in sequence:
    
        if args["no_lat"]:
            if args["no_graph"]:
                res = block.archigrapheme_ar,
            elif args["no_arch"]:
                res = block.grapheme_ar,
            else:
                res = block.grapheme_ar, block.archigrapheme_ar
        elif args["no_ara"]:
            if args["no_graph"]:
                res = block.archigrapheme_lt,
            elif args["no_arch"]:
                res = block.grapheme_lt,
            else:
                res = block.grapheme_lt, block.archigrapheme_lt
        else:
            if args["no_graph"]:
                res = block.archigrapheme_ar, block.archigrapheme_lt
            elif args["no_arch"]:
                res = block.grapheme_ar, block.grapheme_lt
            else:
                res = block.grapheme_ar, block.grapheme_lt, block.archigrapheme_ar, block.archigrapheme_lt

        ind = f"{block.index.sura}:{block.index.verse}:{block.index.word}"
        if block.index.block:
            ind = f"{ind}:{block.index.block}"

        res = res + (ind,)

        yield res



#!/usr/bin/env python3
#
#    models.py
#
# Copyright (c) 2025 Alicia González Martínez
#
################################################

from enum import Enum
from pydantic import BaseModel


class Source(Enum):
    """ Encoding of Quranic text
    """
    TANZIL_SIMPLE = "tanzil-simple"
    TANZIL_UTHMANI = "tanzil-uthmani"
    DECOTYPE = "decotype"

    @classmethod
    def from_str(cls, s: str) -> "Source":
        """Convert string to Source enum member"""
        try:
            return cls(s)
        except ValueError as e:
            raise ValueError(f"Invalid source: {s}. Must be one of {[e.value for e in cls]}") from e

    def get_file(self) -> str:
        """ Get the corresponding Quran file name based on the source
        """
        mapping = {
            Source.TANZIL_SIMPLE: "mushaf_simple.json",
            Source.TANZIL_UTHMANI: "mushaf_uthmani.json",
            Source.DECOTYPE: "mushaf_dt.json"
        }
        return mapping[self]


class Index(BaseModel):
    """ Quranic index.
    """
    sura: int
    verse: int
    word: int
    block: int | None

    @classmethod
    def from_tuple(cls, index: tuple[int, int, int, int]) -> "Index":
        """ Create an Index instance from a tuple.

        Args:
            index_tuple: Quran index.

        """
        if len(index) != 4:
            raise ValueError("Tuple must have exactly four elements")
        
        return cls(sura=index[0], verse=index[1], word=index[2], block=index[3])


    def to_zero_index(self) -> None:
        """ Convert all indexes to 0-index"""
        self.sura -= 1
        self.verse -= 1
        self.word -= 1
        if self.block is not None:
            self.block -= 1


class Block(BaseModel):
    """ Quranic block with corresponding index.
    """
    grapheme_ar: str
    grapheme_lt: str
    archigrapheme_ar: str
    archigrapheme_lt: str 
    index: Index



#!/usr/bin/env python3
#
#    tests_qran.py
#
# Copyright (c) 2025 Alicia González Martínez
#
################################################


import pytest
from qran import get_text, Index

class TestMushaf:
    """Test suite for module1 functionality."""

    def test_general(self):
        """Test basic functionality."""
        
        result = get_text(
            ini_index=Index(sura=1, verse=1, word=1, block=1),
            end_index=Index(sura=1, verse=1, word=1, block=1),
        )

        res = list(result)
        assert len(res) == 1

        graph_ar, graph_lt, arch_ar, arch_lt, ind = res[0]

        assert graph_ar == "بِسْمِ"
        assert graph_lt == "B₁ᵢSᵒMᵢ"
        assert arch_ar == "ٮسم"
        assert arch_lt == "BSM"
        assert ind == "1:1:1"


    def test_overload_index(self):
        """Test basic functionality."""
        
        result = get_text(
            ini_index = (1, 1, 1, 1),
            end_index = (1, 1, 1, 1),
        )

        res = list(result)
        assert len(res) == 1

        graph_ar, graph_lt, arch_ar, arch_lt, ind = res[0]

        assert graph_ar == "بِسْمِ"
        assert graph_lt == "B₁ᵢSᵒMᵢ"
        assert arch_ar == "ٮسم"
        assert arch_lt == "BSM"
        assert ind == "1:1:1"

    def test_overload_source(self):
        """Test basic functionality."""
        
        result = get_text(
            ini_index = (1, 1, 1, 1),
            end_index = (1, 1, 1, 1),
            source = "tanzil-simple",
        )

        res = list(result)
        assert len(res) == 1

        graph_ar, graph_lt, arch_ar, arch_lt, ind = res[0]

        assert graph_ar == "بِسْمِ"
        assert graph_lt == "B₁ᵢSᵒMᵢ"
        assert arch_ar == "ٮسم"
        assert arch_lt == "BSM"
        assert ind == "1:1:1"

    def test_overload_source_err(self):
        """Test basic functionality."""
        
        with pytest.raises(ValueError):
            result = get_text(
                ini_index = (1, 1, 1, 1),
                end_index = (1, 1, 1, 1),
                source = "other-source",
            )
            res = list(result)


class TestMushafBlocks:
    """Test suite for module1 functionality."""

    def test_1111_1111(self):
        """Test index combination"""
        
        result = get_text(
            ini_index=Index(sura=1, verse=1, word=1, block=1),
            end_index=Index(sura=1, verse=1, word=1, block=1),
            args={"blocks": True}
        )

        res = list(result)
        assert res[0][-1] == "1:1:1:1"


    def test_1111_111_(self):
        """Test index combination"""
        
        result = get_text(
            ini_index=Index(sura=1, verse=1, word=1, block=1),
            end_index=Index(sura=1, verse=1, word=1, block=-1),
            args={"blocks": True}
        )

        res = list(result)
        assert res[0][-1] == "1:1:1:1"


    def test_112__112_(self):
        """Test index combination"""
        
        result = get_text(
            ini_index=Index(sura=1, verse=1, word=2, block=0),
            end_index=Index(sura=1, verse=1, word=2, block=-1),
            args={"blocks": True}
        )
        res = list(result)
        assert len(res) == 2
        assert res[0][-1] == "1:1:2:1"
        assert res[1][-1] == "1:1:2:2"


    def test_1622_171_(self):
        """Test index combination"""
        
        result = get_text(
            ini_index=Index(sura=1, verse=6, word=2, block=2),
            end_index=Index(sura=1, verse=7, word=1, block=-1),
            args={"blocks": True}
        )
        res = list(result)
        assert len(res) == 8
        assert res[0][-1] == "1:6:2:2"
        assert res[-1][-1] == "1:7:1:3"
        assert " ".join(r[-2] for r in res) == "LCR A T A LMSBFBM CR A T"




#!/usr/bin/env python3
#
#    qran
#
# standalone entry point for qran
#                         
#    __ _ _ __ __ _ _ __  
#   / _` | '__/ _` | '_ \ 
#  | (_| | | | (_| | | | |
#   \__, |_|  \__,_|_| |_|
#      |_|                
# 
#
# Copyright (c) 2025 Alicia González Martínez
#
####################################################################################

import sys
import textwrap
import orjson as json
from argparse import ArgumentParser, FileType, RawTextHelpFormatter

from qran import __version__
from .util import parse_quran_range
from .models import Source
from .mushaf import get_text


def main():

    parser = ArgumentParser(
        description="Quran structured text API.\n\n" \
        "Unless indicated in parameters, each output token is given in four shapes:\n"\
        "  (1) Arabic graphemic representation\n" \
        "  (2) Latin graphemic representation\n" \
        "  (3) Arabic archgraphemic representation\n" \
        "  (4) Latin graphemic representation\n\n" \
        "Example of usage:\n" \
        "  $ qran 1:7:9:2-2:1:2 --blocks",
        formatter_class=RawTextHelpFormatter,
        epilog="Ya Kabikaj, protect this code from bugs!"
    )
    parser.add_argument(
        "index",
        type=parse_quran_range,
        nargs="?",
        default="1-",
        help=textwrap.dedent(""" \
            Quranic index range to retrieve [DEFAULT WHOLE TEXT]. 
             A complete range must have the following format:
                ini_sura:ini_verse:ini_word:ini_block-end_sura:end_verse:end_word:end_block
             Both beginning and end indexes are inclusive. All indexes are optional.
        """))
    parser.add_argument(
        "--source", "-s",
        choices=[s.value for s in Source],
        default=Source.TANZIL_SIMPLE.value,
        help="Quran encoding"
    )
    parser.add_argument(
        "--blocks",
        action="store_true",
        help="retrieve text as letterblocks instead of words"
    )
    script = parser.add_mutually_exclusive_group()
    script.add_argument(
        "--no_lat",
        action="store_true",
        help="omit Latin traslineration in output"
    )
    script.add_argument(
        "--no_ara",
        action="store_true",
        help="omit Arabic script in output"
    )
    layer = parser.add_mutually_exclusive_group()
    layer.add_argument(
        "--no_arch",
        action="store_true",
        help="omit archigraphemic representations in output"
    )
    layer.add_argument(
        "--no_graph",
        action="store_true",
        help="omit graphemic representations in output"
    )
    parser.add_argument(
        "--sep",
        default="\t",
        help="field separator for text output [DEFAULT \\t]"
    )
    parser.add_argument(
        "--json",
        action="store_true",
        help="print output in json instead of plain text"
    )
    parser.add_argument(
        "--out",
        type=FileType("w"),
        help="write output in file instead of stdin"
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="prints the program version number and exits successfully"
    )
    args = parser.parse_args()

    try:
        tokens = get_text(
            ini_index=args.index[0],
            end_index=args.index[1],
            source=Source.from_str(args.source),
            args={
                "blocks": args.blocks,
                "no_lat": args.no_lat,
                "no_ara": args.no_ara,
                "no_graph": args.no_graph,
                "no_arch": args.no_arch,
            }
        )

        out = args.out or sys.stdout

        if args.json:
            data = [{"tok": tok[:-1], "ind": tok[-1]} for tok in tokens]
            out.write(json.dumps(data).decode("utf-8"))

        else:
            for tok in tokens:
                print(args.sep.join(tok), file=out)
    
    except FileNotFoundError as err:
        print(err, file=sys.stderr)

    except (KeyboardInterrupt, BrokenPipeError, IOError):
        pass

    sys.stderr.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
#    update_data.py
#
# Script to fix errors in Quranic source data files
#
# check charset:
#   $ qran --source tanzil-uthmani | awk '{print $2}' | grep -o . | sort -u
#
# chexk output:
#   $ cat mushaf_uthmani-updated.json | jq -r '.blocks | flatten[]' | grep -o . | sort -u
#
# Copyright (c) 2025 Alicia González Martínez
#
# usage:
#   $ python3 update_data.py mushaf_uthmani.json mushaf_uthmani-updated.json
#     --latin_graph "/ـ//"
#
#########################################################################################


import re
import sys
import json
from typing import Pattern
from argparse import ArgumentParser, FileType, ArgumentTypeError



def _parse_sub(value: str) -> tuple[Pattern, str]:
    """ parse arg with a pattern and replacement modification

    Args:
        arg: argument to parse

    Return:
        regex and replacement.

    Raise:
        ValueError: if value is ill-formed.

    """
    try:
        if value[0] == value[-1] == "/":
            value = value[1:-1]
            pat, repl = value.split("/")
            return re.compile(pat), repl
        raise ArgumentTypeError
    except ValueError:
        raise ArgumentTypeError


def main():
    parser = ArgumentParser(
    	description="Script to fix errors in Quranic source data files"
    )
    parser.add_argument(
        "infile",
        type=FileType("r"),
        nargs="?",
        default=sys.stdin,
        help="Quran data file"
    )
    parser.add_argument(
        "outfile",
        type=FileType("w"),
        nargs="?",
        default=sys.stdout,
        help="modified file"
    )
    parser.add_argument(
        "--latin_graph",
        type=_parse_sub,
        help="modify field with regex, format: /pattern/replacement/"
    )
    args = parser.parse_args()

    data = json.load(args.infile)

    latin_graph_regex, latin_graph_repl = args.latin_graph

    for iblock in range(len(data["blocks"])):
        arabic_graph, latin_graph, arabic_arch, latin_arch = data["blocks"][iblock]
        data["blocks"][iblock][1] = latin_graph_regex.sub(latin_graph_repl, latin_graph)

    # for isura in range(len(data["indexes"])):
    #     sura = data["indexes"][isura]
    #     for ivers in range(len(data["indexes"][isura])):
    #         vers = data["indexes"][isura][ivers]
    #         for iword in range(len(data["indexes"][isura][ivers])):
    #             word = data["indexes"][isura][ivers][iword]
    #             for iblock in range(len(data["indexes"][isura][ivers][iword])):
    #                 block = data["indexes"][isura][ivers][iword][iblock]

    json.dump(data, args.outfile)


if __name__ == "__main__":
    main()

